# -*- encoding : utf-8 -*-
require 'grit'
require 'grit_ext/status' # add ignored support
require 'thor'
require 'fileutils'

module GazelleStyleguide
  # Command line tool for managing linter running and pre-commits
  class CLI < Thor
    GENERIC_CHECKS = 'jshint, merge_conflict, json, yaml, coffeelint, scss_lint'
    JAVA_CHECKS = 'checkstyle'
    RUBY_CHECKS = 'rubocop, rspec_focus'

    desc 'init', 'Sets up pre-commit hooks to check style.'
    def init
      setup_git_config
      install_pre_commit
      setup_houndci
    end

    desc 'lint [flags|FILES]', 'Runs linters'
    option :all, aliases: :a
    option :changed, aliases: :c
    option :staged, aliases: :s
    def lint(*files)
      require 'pre-commit'
      files = lint_files(files, options)

      PreCommit::Runner.new($stderr, files).run
    end

    desc 'fix [flags|FILES]', 'Runs rubocop fix against ruby files'
    option :all, aliases: :a
    option :changed, aliases: :c
    option :staged, aliases: :s
    option :debug, aliases: :d
    def fix(*files)
      require 'rubocop'
      files = lint_files(files, options).select{|file| file =~ /\.rb|Gemfile|Rakefile|\.task|\.rake/}
      args = ['-a', '-c', GazelleStyleguide.config_for('rubocop.yml')] + files
      args.unshift('-d') if options[:debug]

      Rubocop::CLI.new.run(args)
    end

  private

    def install_pre_commit
      require 'pre-commit/installer'

      PreCommit::Installer.new.install
    end

    def setup_git_config
      checks = GENERIC_CHECKS + ', ' + (ruby? ? RUBY_CHECKS : JAVA_CHECKS)
      repo.config['pre-commit.checks'] = checks
      repo.config['pre-commit.jshint.config'] = GazelleStyleguide.config_for('jshintrc.json')
      repo.config['pre-commit.rubocop.config'] = GazelleStyleguide.config_for('rubocop.yml')
      repo.config['pre-commit.coffeelint.config'] = GazelleStyleguide.config_for('coffeelint.json')
      repo.config['pre-commit.checkstyle.config'] = GazelleStyleguide.config_for('sun_checks.xml')
    end

    def setup_houndci
      return unless ruby?

      FileUtils.cp(GazelleStyleguide.config_for('rubocop.yml'), project_root.join('.hound.yml'))
    end

    def project_root
      @project_root ||= Pathname.new('.').ascend do |path|
        return path if path.join('.git').exist?
      end
    end

    def ruby?
      project_root.join('Gemfile').exist?
    end

    def lint_files(files, options)
      if files.any?
        files
      elsif options[:all]
        Dir['./**/*'].select{|f| File.file?(f)}
      elsif options[:staged]
        staged_files
      else
        changed_files
      end
    end

    def staged_files
      repo.status.changed.keys + repo.status.added.keys
    end

    def changed_files
      repo.status.untracked.keys + staged_files
    end

    def repo
      @repo ||= Grit::Repo.new('.')
    end
  end
end
